import unittest
import os
import copy

import pymongo
import alfred

from alfred3_dbtools import mongotools


class TestMongoDBConnector(unittest.TestCase):
    def setUp(self):
        self._host = os.environ.get("MONGODB_TESTHOST")
        self._port = int(os.environ.get("MONGODB_TESTPORT"))
        self._user1 = os.environ.get("MONGODB_TESTUSER")
        self._pw1 = os.environ.get("MONGODB_TESTPW")
        self._db = os.environ.get("MONGODB_TESTDB")
        self._col = os.environ.get("MONGODB_TESTCOL")
        self._user2 = os.environ.get("MONGODB_TESTUSER2")
        self._pw2 = os.environ.get("MONGODB_TESTPW2")
        self._auth_source = os.environ.get("MONGODB_TEST2AUTH")

    def test_minimal_connect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=self._port,
            username=self._user1,
            password=self._pw1,
            database=self._db,
        )
        ret = con.connect()
        con.connect()
        self.assertTrue(con.connected)
        self.assertIsInstance(ret, pymongo.database.Database)
        self.assertIsInstance(con.db, pymongo.database.Database)

    def test_auth_source_connect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=self._port,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            auth_source=self._auth_source,
        )
        ret = con.connect()
        con.connect()
        self.assertTrue(con.connected)
        self.assertIsInstance(ret, pymongo.database.Database)
        self.assertIsInstance(con.db, pymongo.database.Database)

    def test_collection_connect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=self._port,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            collection=self._col,
            auth_source=self._auth_source,
        )
        ret = con.connect()
        self.assertTrue(con.connected)
        self.assertIsInstance(ret, pymongo.collection.Collection)
        self.assertIsInstance(con.db, pymongo.collection.Collection)

    def test_autoconnect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=self._port,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            collection=self._col,
            auth_source=self._auth_source,
        )
        self.assertFalse(con.connected)
        col = con.db
        self.assertTrue(con.connected)
        self.assertIsInstance(col, pymongo.collection.Collection)

    def test_disconnect(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=self._port,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            collection=self._col,
            auth_source=self._auth_source,
        )
        col = con.connect()
        self.assertIsInstance(col, pymongo.collection.Collection)
        con.disconnect()

    def test_insert_find_remove(self):
        con = mongotools.MongoDBConnector(
            host=self._host,
            port=self._port,
            username=self._user2,
            password=self._pw2,
            database=self._db,
            collection=self._col,
            auth_source=self._auth_source,
        )
        col = con.connect()
        test_document = {"test": "test"}

        col.insert_one(test_document)
        query = col.find_one(test_document)
        self.assertIsNotNone(query)

        col.delete_one(test_document)
        delquery = col.find_one(test_document)
        self.assertIsNone(delquery)


class TestExpMongoDBConnector(unittest.TestCase):
    def setUp(self):
        self._host = os.environ.get("MONGODB_TESTHOST")
        self._port = int(os.environ.get("MONGODB_TESTPORT"))
        self._user = os.environ.get("MONGODB_TESTUSER2")
        self._pw = os.environ.get("MONGODB_TESTPW2")
        self._db = os.environ.get("MONGODB_TESTDB")
        self._col = os.environ.get("MONGODB_TESTCOL")
        self._auth_source = os.environ.get("MONGODB_TEST2AUTH")

        self.exp = alfred.Experiment()

        self.agent = alfred.saving_agent.MongoSavingAgent(
            host=self._host + ":" + str(self._port),
            database=self._db,
            collection=self._col,
            user=self._user,
            password=self._pw,
            experiment=self.exp,
            auth_source=self._auth_source,
            use_ssl=False,
            ca_file_path=None,
            activation_level=10
        )

        self.exp.saving_agent_controller.add_saving_agent(self.agent)
        self.connector = mongotools.ExpMongoDBConnector(self.exp)

    def test_connect(self):
        col = self.connector.connect()
        self.assertIsInstance(col, pymongo.collection.Collection)
        self.assertTrue(self.connector.connected)
    
    def test_autoconnect(self):
        self.assertFalse(self.connector.connected)

        col = self.connector.db
        self.assertIsInstance(col, pymongo.collection.Collection)
        self.assertTrue(col.connected)
    
    def test_change(self):
        
        self.connector.connect()
        agent = self.connector._agents[0]
        agent.activation_level = 9

        self.assertEqual(agent.activation_level, 9)
        self.assertEqual(self.exp.saving_agent_controller._agents[1].activation_level, 10)
    
    def test_multi_agent_handling(self):
        agent2 = copy.copy(self.agent)
        agent3 = copy.copy(self.agent)
        agent4 = copy.copy(self.agent)
        
        agent2.activation_level = 10
        agent3.activation_level = 9
        agent4.activation_level = 9

        # additional agent, same level
        self.exp.saving_agent_controller.add_saving_agent(agent2)
        connector1 = mongotools.ExpMongoDBConnector(self.exp)
        self.assertRaises(ValueError, self.connector.connect)
        self.assertRaises(ValueError, connector1.connect)

        # third agent, lower level
        self.exp.saving_agent_controller.add_saving_agent(agent3)
        col = self.connector.connect()
        self.assertTrue(self.connector.connected)
        self.assertIsInstance(col, pymongo.collection.Collection)

        # fourth agent, same level as third
        self.exp.saving_agent_controller.add_saving_agent(agent4)
        connector2 = mongotools.ExpMongoDBConnector(self.exp)
        self.assertRaises(ValueError, self.connector.connect)
        self.assertRaises(ValueError, connector2.connect)
    
    def test_insert_find_remove(self):
        col = self.connector.connect()

        test_document = {"test": "test"}

        col.insert_one(test_document)
        query = col.find_one(test_document)
        self.assertIsNotNone(query)

        col.delete_one(test_document)
        delquery = col.find_one(test_document)
        self.assertIsNone(delquery)
    
    def test_list_agents(self):
        agent2 = copy.copy(self.agent)
        agent3 = copy.copy(self.agent)
        
        agent2.activation_level = 10
        agent3.activation_level = 9

        self.exp.saving_agent_controller.add_saving_agent(agent2)
        self.exp.saving_agent_controller.add_saving_agent(agent3)

        connector = mongotools.ExpMongoDBConnector(self.exp)
        agents_list = connector.list_agents

        self.assertEqual(len(agents_list), 3)
    
    def test_client(self):
        self.assertFalse(self.connector.connected)
        self.connector.connect()
        self.assertIsInstance(self.connector._client, pymongo.MongoClient)
    
    def test_input(self):
        with self.assertRaises(ValueError):
            mongotools.ExpMongoDBConnector("test")
    
    def test_no_mongo_agent(self):
        exp = alfred.Experiment()
        with self.assertRaises(ValueError):
            connector = mongotools.ExpMongoDBConnector(exp)
            connector.connect()


if __name__ == "__main__":
    unittest.main()



"""Tools for interacting with MongoDB databases."""

import copy

from pymongo import MongoClient
import alfred


class MongoDBConnector:
    """Connect to a MongoDB database or collection.
    
    :param str host: Hostname or IP address.
    :param int port: Port number on which to connect.
    :param str username: Username for authentication.
    :param str password: Password for authentication.
    :param str database: Database to which to the client will connect.
    :param str collection: Collection inside the specified database to which the client will connect. If left empty, the client will connect directly to the database.
    :param str auth_source: Database to authenticate on. Defaults to "admin", the MongoDB default.
    :param bool ssl: If ``True``, create a connection to the server using Transport Layer Security (TLS/SSL).
    :param str ca_file: Filepath to a file containing a single or a bundle of “certification authority” certificates, which are used to validate certificates passed from the other end of the connection. Implies ``tls=True``. Defaults to ``None``.
    """

    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        database: str,
        collection: str=None,
        auth_source: str="admin",
        ssl: bool=False,
        ca_file: str=None,
    ):
        """Constructor method."""

        self._host = host
        self._port = port
        self._db_name = database
        self._collection = collection
        self._username = username
        self._password = password
        self._auth_source = auth_source
        self._ssl = ssl
        self._ca_file = ca_file

        self._client = None
        self._db = None

        self.connected = False

    def connect(self):
        """Establish the connection the MongoDB.

        If a collection was specified upon initialisation, the DBConnector will connect to this collection. Else, it will connect to the specified database.

        :return: The database or, if specified, collection given upon initialisation.
        """

        self._client = MongoClient(
            host=self._host,
            port=self._port,
            username=self._username,
            password=self._password,
            authSource=self._auth_source,
            tls=self._ssl,
            tlsCAFile=self._ca_file,
        )

        db = self._client[self._db_name]

        if self._collection:
            db = db[self._collection]

        self._db = db
        self.connected = True

        return self._db

    def disconnect(self):
        """Close the connection to the database."""

        self._client.close()
        self.connected = False

    @property
    def db(self):
        """Return the database or, if specified, collection given upon initialisation."""

        if self._db:
            return self._db
        else:
            return self.connect()


class ExpMongoDBConnector:
    """Connect to an alfred experiment's MongoDB collection.
    
    Connects to the same database and collection that the given alfred experiment uses to save its data. If multiple MongoSavingAgents are attached to the experiment, the ExpMongoDBConnector will connect to the MongoSavingAgent with the lowest activation level.

    :param experiment: An instance of ``alfred.Experiment``.
    :type experiment: class: `alfred.Experiment`
    :raise: ValueError if not initialised with an alfred experiment.
    """

    def __init__(self, experiment):
        """Constructor method."""

        self._exp = experiment
        self._agents = None
        
        self._client = None
        self._collection = None
        
        self.connected = False

        if not isinstance(self._exp, alfred.Experiment):
            raise ValueError("The input must be an instance of alfred.Experiment.")

    def _gather_agents(self):
        """Collect all MongoSavingAgents from the provided alfred experiment, sorted by activation level (lowest first)."""

        self._agents = []
        for agent in self._exp.saving_agent_controller._agents:
            if isinstance(agent, alfred.saving_agent.MongoSavingAgent):
                self._agents.append(copy.copy(agent))
        if not self._agents:
            raise ValueError("Your experiment needs at least one MongoSavingAgent for ExpMongoDBConnector to work.")
        self._agents.sort(key=lambda x: x.activation_level)
        
    def connect(self):
        """Establish a connection to the experiment's MongoDB with the lowest activation level.
        
        :return: The MongoClient collection from the MongoSavingAgent with lowest activation level.
        """        

        self._gather_agents()        
        
        if len(self._agents) > 1 and self._agents[0].activation_level == self._agents[1].activation_level:
            raise ValueError("There are two or more MongoSavingAgents with the highest activation level.")

        self._client = self._agents[0]._mc
        self._collection = self._agents[0]._col
        self.connected = True

        return self._collection
    
    def disconnect(self):
        """Close the connection to the database."""

        if self.connected:
            self._client.close()
            self.connected = False
    
    @property
    def db(self):
        """Return the experiment's MongoDB collection."""

        if self._collection:
            return self._collection
        else:
            return self.connect()
    
    @property
    def list_agents(self):
        """Return a list of all MongoSavingAgents belonging to the given alfred experiment."""
        if self._agents:
            return self._agents
        else:
            self._gather_agents()
            return self._agents



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="alfred3_dbtools",
    version="0.1.1",
    author="Christian Treffenstädt, Johannes Brachem",
    author_email="treffenstaedt@psych.uni-goettingen.de",
    description="A package that provides tools for interacting with databases when working with alfred3 experiments.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ctreffe/alfred3-dbtools",
    packages=setuptools.find_packages(),
    install_requires=["alfred>=1.0.0", "pymongo>=3.10"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)